/*
 *	System Atari functions
 *
 *	Patrice Mandin
 *	Xavier Joubert
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>

#include <stdarg.h>
#include <sys/time.h>
#include <unistd.h>

#include <mint/osbind.h>
#include <sysvars.h>

#include "doomdef.h"
#include "m_misc.h"
#include "i_video.h"
#include "i_sound.h"
#include "i_net.h"

#include "d_net.h"
#include "g_game.h"
#include "m_argv.h"

#include "i_cookies.h"
#include "in_ikbd.h"
#include "in_joypad.h"
#include "in_bios.h"
#include "vid_xbios.h"
#include "vid_vdi.h"
#include "vid_gem.h"
#include "snd_atari.h"
#include "snd_server.h"
#include "net_unix.h"
#include "net_sting.h"

#ifdef __GNUG__
#pragma implementation "i_system.h"
#endif
#include "i_system.h"

enum {
	MX_STRAM=0,
	MX_TTRAM,
	MX_PREFSTRAM,
	MX_PREFTTRAM
};

int	mb_used = 8;	/*  8 Mo par dfaut */
int mxalloc_present = 0;	/* Mxalloc system call present ? */

void I_NullSysFunc(void);

void
I_Tactile
( int	on,
  int	off,
  int	total )
{
  /*  UNUSED. */
  on = off = total = 0;
}

ticcmd_t	emptycmd;
ticcmd_t*	I_BaseTiccmd(void)
{
    return &emptycmd;
}

int atari_HeapSize=8<<20;	/* 8 Mo par defaut */

int  I_GetHeapSize (void)
{
    return mb_used*1024*1024;
}

byte* I_ZoneBase (int*	size)
{
	if (mxalloc_present)
		*size=Mxalloc(-1,MX_PREFTTRAM)-131072;
	else
		*size=Malloc(-1)-131072;

	if (*size>atari_HeapSize)
		*size=atari_HeapSize;

	printf(" %d Kbytes allocated for zone\n",(*size)>>10);

    return (byte *) malloc (*size);
}



/*  */
/*  I_GetTime */
/*  returns time in 1/70th second tics */
/*  */

int  I_GetTime (void)
{
    static int		basetime=0;
	void *oldpile;
	int atari_curtime;

	oldpile=(void *)Super(0);
	atari_curtime=*((volatile long *)_hz_200);
	Super(oldpile);

	if (!basetime)
		basetime = atari_curtime;

	return ((atari_curtime-basetime)*TICRATE)/200;	/* Return 35hz value */
}


/*  */
/*  I_Init */
/*  */
void I_Init (void)
{
	I_InitSound();
	/* I_InitGraphics(); */
	I_InitControl();

	signal(SIGINT, (void (*)(int)) I_Quit);
}

/*  */
/*  I_Quit */
/*  */
void I_Quit (void)
{
    D_QuitNetGame ();
	I_ShutdownNetwork();
	I_ShutdownControl();
    I_ShutdownSound();
    I_ShutdownMusic();
    M_SaveDefaults ();
    I_ShutdownGraphics();
    exit(0);
}

void I_WaitVBL(int count)
{
	int i;

	for (i=0;i<count;i++)
		Vsync();
}

byte*	I_AllocLow(int length)
{
    byte*	mem;
        
    mem = (byte *)malloc (length);
    memset (mem,0,length);
    return mem;
}


/*  */
/*  I_Error */
/*  */
extern boolean demorecording;

void I_Error (char *error, ...)
{
    va_list	argptr;

    /*  Message first. */
    va_start (argptr,error);
    fprintf (stderr, "Error: ");
    vfprintf (stderr,error,argptr);
    fprintf (stderr, "\n");
    va_end (argptr);

    fflush( stderr );

    /*  Shutdown. Here might be other errors. */
    if (demorecording)
	G_CheckDemoStatus();

    D_QuitNetGame ();
	I_ShutdownNetwork();
	I_ShutdownControl();
	I_ShutdownSound();
	I_ShutdownMusic();
    I_ShutdownGraphics();
    
    exit(-1);
}

void I_StartTic(void)
{
	I_UpdateKeyboard();
	I_UpdateMouse();
	I_UpdateJoystick();
	I_UpdateJoypad();
}

/* ----------------------------------------------------------------- */
/*                 Dynamic system dependent functions                */
/* ----------------------------------------------------------------- */

typedef struct {
	char *name;	/* Routine name */
	char *param;	/* Parameter to force the use of this routine */
	int number;	/* Number of the routine */
} routlist_t;

enum {
	SYSROUT_OFF =0,
	SYSROUT_ON,
	SYSROUT_6301,
	SYSROUT_BIOS,
	SYSROUT_SNDSERV,
	SYSROUT_SNDDMA8,
	SYSROUT_XBIOS,
	SYSROUT_VDI,
	SYSROUT_GEM,
	SYSROUT_SOCKETS,
	SYSROUT_STING
};

routlist_t system_rout[]={
	{"Off",		"off",		SYSROUT_OFF},
	{"On",		"on",		SYSROUT_ON},
	{"Ikbd 6301",	"ikbd",		SYSROUT_6301},
	{"Bios",	"bios",		SYSROUT_BIOS},
	{"Sound server","sndserver",	SYSROUT_SNDSERV},
	{"8 bits Dma",	"dma8",		SYSROUT_SNDDMA8},
	{"Xbios",	"xbios",	SYSROUT_XBIOS},
	{"Vdi",		"vdi",		SYSROUT_VDI},
	{"Gem",		"gem",		SYSROUT_GEM},
	{"Unix sockets","unix",		SYSROUT_SOCKETS},
	{"Sting",	"sting",	SYSROUT_STING}
};

void I_SysInit(void)
{
	int p;

	int numkbd,nummouse,numjoystick;
	int numjoypad;
	int numsound,numvideo;
	int numnetwork;

	/* --- Initialize default system functions --- */

	/* controllers */

	I_InitControl = I_NullSysFunc; /* Init controller functions */
	I_ShutdownControl = I_NullSysFunc; /* Shutdown controller functions */

	I_UpdateKeyboard = I_NullSysFunc;
	numkbd = SYSROUT_OFF;
	I_UpdateMouse = I_NullSysFunc;
	nummouse = SYSROUT_OFF;
	I_UpdateJoystick = I_NullSysFunc;
	numjoystick = SYSROUT_OFF;
	I_UpdateJoypad = I_NullSysFunc;
	numjoypad = SYSROUT_OFF;

	/* sound */

	I_SndInit = I_NullSysFunc;
	I_SndUpdate = I_NullSysFunc;
	I_InitSound = I_NullSysFunc;
	I_ShutdownSound = I_NullSysFunc;
	I_StartSound = I_NullSysFunc;
	numsound = SYSROUT_OFF;

	/* video */

	I_VidInit = I_NullSysFunc;
	I_VidUpdate = I_NullSysFunc;
	I_SetPalette256 = I_NullSysFunc;
	I_InitGraphics = I_InitGraphics_null;
	I_ShutdownGraphics = I_NullSysFunc;
	I_SetMainScreen = I_NullSysFunc;
	numvideo = SYSROUT_OFF;

	/* network */

	I_InitNetwork = I_InitNetwork_unix;
	I_ShutdownNetwork = I_ShutdownNetwork_unix;
	numnetwork = SYSROUT_SOCKETS;

	/* test for Mxalloc */

	mxalloc_present = (((Sversion()&0xFF)>=0x01)|(Sversion()>=0x1900));

	/* --- Initialize default ATARI functions --- */

	I_ScanCookies();
	
	/* controllers */
	if ((cookie_mch == MCH_ST) || (cookie_mch == MCH_STE)
		|| (cookie_mch == MCH_TT) || (cookie_mch == MCH_F30))
	{
		I_InitControl = I_InitControl_6301;
		I_ShutdownControl = I_ShutdownControl_6301;

		I_UpdateKeyboard = I_UpdateKeyboard_6301;
		numkbd=SYSROUT_6301;
		I_UpdateMouse = I_UpdateMouse_6301;
		nummouse=SYSROUT_6301;
		I_UpdateJoystick = I_UpdateJoystick_6301;
		numjoystick=SYSROUT_6301;
	}
	else
	{
		I_InitControl = I_InitControl_bios;
		I_UpdateKeyboard = I_UpdateKeyboard_bios;
		numkbd=SYSROUT_BIOS;
	}

	if ((cookie_mch == MCH_STE) || (cookie_mch == MCH_F30))
	{
		I_UpdateJoypad = I_UpdateJoypad_hardware;
		numjoypad=SYSROUT_ON;
	}

	/* video */
	switch (cookie_vdo) {
		case VDO_ST:
		case VDO_STE:
		case VDO_TT:
		case VDO_F30:
			/* Xbios display */
			I_VidInit = I_VidInit_xbios;
			I_VidUpdate = I_VidUpdate_xbios;
			I_SetPalette256 = I_SetPalette256_xbios;
			I_InitGraphics = I_InitGraphics_xbios;
			I_ShutdownGraphics = I_ShutdownGraphics_xbios;
			I_SetMainScreen = I_NullSysFunc;
			numvideo=SYSROUT_XBIOS;
			break;
		default:
			/* Vdi display */
			I_VidInit = I_VidInit_vdi;
			I_VidUpdate = I_VidUpdate_vdi;
			I_SetPalette256 = I_SetPalette256_vdi;
			I_InitGraphics = I_InitGraphics_vdi;
			I_ShutdownGraphics = I_ShutdownGraphics_vdi;
			I_SetMainScreen = I_SetMainScreen_vdi;
			numvideo=SYSROUT_VDI;
			break;
	}

	/* sound */
	if (cookie_snd & (1<<SND_DMA8)) {
		I_SndInit = I_SndInit_atari;
		I_SndUpdate = I_NullSysFunc;
		I_InitSound = I_InitSound_atari;
		I_ShutdownSound = I_ShutdownSound_atari;
		I_StartSound = I_StartSound_atari;
		numsound=SYSROUT_SNDDMA8;
	}

	/* --- Initialise forced functions --- */

	p=M_CheckParm("-keyboard");
	if (p && p < myargc-1 )
	{
		if (!strcmp(system_rout[SYSROUT_OFF].param,myargv[p+1])) {
			I_UpdateKeyboard = I_NullSysFunc;
			numkbd=SYSROUT_OFF;
		}		
		if (!strcmp(system_rout[SYSROUT_BIOS].param,myargv[p+1]))
		{
			/* Set bios routines */
			I_InitControl=I_InitControl_bios;
			I_ShutdownControl=I_NullSysFunc;

			I_UpdateKeyboard=I_UpdateKeyboard_bios;
			numkbd=SYSROUT_BIOS;
			I_UpdateMouse=I_NullSysFunc;
			nummouse=SYSROUT_OFF;
			I_UpdateJoystick=I_NullSysFunc;
			numjoystick=SYSROUT_OFF;
		}
	}

	p=M_CheckParm("-mouse");
	if (p && p < myargc-1 )
	{
		if (!strcmp(system_rout[SYSROUT_OFF].param,myargv[p+1])) {
			I_UpdateMouse = I_NullSysFunc;
			nummouse=SYSROUT_OFF;
		}		
	}

	p=M_CheckParm("-joystick");
	if (p && p < myargc-1 )
	{
		if (!strcmp(system_rout[SYSROUT_OFF].param,myargv[p+1])) {
			I_UpdateJoystick = I_NullSysFunc;
			numjoystick=SYSROUT_OFF;
		}		
	}

	/* disable IKBD routines if keyboard disabled */
	if ((numkbd == SYSROUT_OFF) && (I_InitControl == I_InitControl_6301))
	{
		I_InitControl = I_NullSysFunc;
		I_ShutdownControl = I_NullSysFunc;

		I_UpdateKeyboard = I_NullSysFunc;
		numkbd =SYSROUT_OFF;
		I_UpdateMouse = I_NullSysFunc;
		nummouse=SYSROUT_OFF;
		I_UpdateJoystick = I_NullSysFunc;
		numjoystick=SYSROUT_OFF;
	}

	p=M_CheckParm("-joypad");
	if (p && p < myargc-1 )
	{
		if (!strcmp(system_rout[SYSROUT_OFF].param,myargv[p+1])) {
			I_UpdateJoypad = I_NullSysFunc;
			numjoypad=SYSROUT_OFF;
		}		
	}

	p=M_CheckParm("-sound");
	if (p && p < myargc-1 )
	{
		if (!strcmp(system_rout[SYSROUT_OFF].param,myargv[p+1])) {
			I_SndInit = I_NullSysFunc;
			I_SndUpdate = I_NullSysFunc;
			I_InitSound = I_NullSysFunc;
			I_ShutdownSound = I_NullSysFunc;
			numsound=SYSROUT_OFF;
		}		
		if (!strcmp(system_rout[SYSROUT_SNDSERV].param,myargv[p+1])) {
			I_SndInit = I_NullSysFunc;
			I_SndUpdate = I_NullSysFunc;
			I_InitSound = I_InitSound_sndserver;
			I_ShutdownSound = I_ShutdownSound_sndserver;
			I_StartSound = I_StartSound_sndserver;
			numsound=SYSROUT_SNDSERV;
		}		
		if (!strcmp(system_rout[SYSROUT_SNDDMA8].param,myargv[p+1])) {
			I_SndInit = I_SndInit_atari;
			I_SndUpdate = I_NullSysFunc;
			I_InitSound = I_InitSound_atari;
			I_ShutdownSound = I_ShutdownSound_atari;
			I_StartSound = I_StartSound_atari;
			numsound=SYSROUT_SNDDMA8;
		}		
	}

	p=M_CheckParm("-video");
	if (p && p < myargc-1 )
	{
		if (!strcmp(system_rout[SYSROUT_OFF].param,myargv[p+1])) {
			I_VidInit = I_NullSysFunc;
			I_VidUpdate = I_NullSysFunc;
			I_SetPalette256 = I_NullSysFunc;
			I_InitGraphics = I_InitGraphics_null;
			I_ShutdownGraphics = I_NullSysFunc;
			I_SetMainScreen = I_NullSysFunc;
			numvideo=SYSROUT_OFF;
		}		
		if (!strcmp(system_rout[SYSROUT_GEM].param,myargv[p+1])) {
			I_VidInit = I_VidInit_gem;
			I_VidUpdate = I_VidUpdate_gem;
			I_SetPalette256 = I_SetPalette256_gem;
			I_InitGraphics = I_InitGraphics_gem;
			I_ShutdownGraphics = I_ShutdownGraphics_gem;
			I_SetMainScreen = I_SetMainScreen_gem;
			numvideo=SYSROUT_GEM;
			/* GEM gets control for input */
			I_InitControl = I_InitControl_gem;
			I_ShutdownControl = I_NullSysFunc;
			I_UpdateKeyboard = I_UpdateKeyboard_gem;
			numkbd =SYSROUT_GEM;
			I_UpdateMouse = I_NullSysFunc;
			nummouse=SYSROUT_OFF;
			I_UpdateJoystick = I_NullSysFunc;
			numjoystick=SYSROUT_OFF;
		}		
		if (!strcmp(system_rout[SYSROUT_VDI].param,myargv[p+1])) {
			I_VidInit = I_VidInit_vdi;
			I_VidUpdate = I_VidUpdate_vdi;
			I_SetPalette256 = I_SetPalette256_vdi;
			I_InitGraphics = I_InitGraphics_vdi;
			I_ShutdownGraphics = I_ShutdownGraphics_vdi;
			I_SetMainScreen = I_SetMainScreen_vdi;
			numvideo=SYSROUT_VDI;
		}		
		if (!strcmp(system_rout[SYSROUT_XBIOS].param,myargv[p+1])) {
			I_VidInit = I_VidInit_xbios;
			I_VidUpdate = I_VidUpdate_xbios;
			I_SetPalette256 = I_SetPalette256_xbios;
			I_InitGraphics = I_InitGraphics_xbios;
			I_ShutdownGraphics = I_ShutdownGraphics_xbios;
			I_SetMainScreen = I_NullSysFunc;
			numvideo=SYSROUT_XBIOS;
		}		
	}

	p=M_CheckParm("-network");
	if (p && p < myargc-1 )
	{
		if (!strcmp(system_rout[SYSROUT_STING].param,myargv[p+1])) {
			I_InitNetwork = I_InitNetwork_sting;
			I_ShutdownNetwork = I_ShutdownNetwork_sting;
			numnetwork=SYSROUT_STING;
		}		
		if (!strcmp(system_rout[SYSROUT_SOCKETS].param,myargv[p+1])) {
			I_InitNetwork = I_InitNetwork_unix;
			I_ShutdownNetwork = I_ShutdownNetwork_unix;
			numnetwork=SYSROUT_SOCKETS;
		}		
	}

	/* Display routines names */

	printf("Keyboard.: %s\n",system_rout[numkbd].name);
	printf("Mouse....: %s\n",system_rout[nummouse].name);
	printf("Joystick.: %s\n",system_rout[numjoystick].name);
	printf("Joypad...: %s\n",system_rout[numjoypad].name);
	printf("Sound....: %s\n",system_rout[numsound].name);
	printf("Video....: %s\n",system_rout[numvideo].name);
	printf("Network..: %s\n\n",system_rout[numnetwork].name);

	/* First initializations */

	I_VidInit();
	I_SndInit();
}

/* --- Default system dynamic function = Null */

void I_NullSysFunc(void)
{
}

/* --- Controllers : update (keyboard,mouse,joystick,joypads,etc... */

/* void (*I_StartTic)(void); */

void (*I_InitControl) (void);
void (*I_ShutdownControl) (void);

void (*I_UpdateKeyboard) (void);
void (*I_UpdateMouse) (void);
void (*I_UpdateJoystick) (void);
void (*I_UpdateJoypad) (void);

/* --- Sound */

void (*I_SndInit)(void);		/* Initializations required for sound */
void (*I_SndUpdate)(void);		/* Update sound */
void (*I_InitSound)(void);		/* Initialize sound */
void (*I_ShutdownSound)(void);		/* Shutdown sound */

/* --- Music */

/* --- Video */

void (*I_VidInit)(void);		/* Initializations required for video */
void (*I_VidUpdate)(void);		/* Render the final screen */
void (*I_SetPalette256)(byte *palette); /* set a 256 colours palette */
void (*I_InitGraphics)(void);		/* Init video */
void (*I_ShutdownGraphics)(void);	/* Shutdown video */
void (*I_SetMainScreen)(void);		/* Set final screen pointer */
